﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;
using System.Configuration;
using System.Threading;

namespace XmlProtocolSampleApp
{
    public class ClientSocketIO : XmlProtocolSampleApp.XmlProtocolController.SocketIO
    {
        /// <summary>
        /// IP endpoint used to define the IP of the Baxi Agent
        /// </summary>
        private static IPEndPoint _remoteEP;

        private const int _retryConnectCount = 2;

        /// <summary>
        /// Array of bytes for incomming messages
        /// </summary>
        private byte[] _bytesReceived = new byte[5120];
        private static Socket _socket;

        public string RemoteEndPoint { get { return _socket.RemoteEndPoint.ToString(); } }
        public bool IsConnected { get { return _socket.Connected; } }

        /// <summary>
        /// Eventhandler for stopping keep-alive messages from being sent.
        /// </summary>
        public EventHandler StopSendingEmptyMessage;

        public ClientSocketIO(string baxiAgentIpAddress, int port)
        {
            IPHostEntry _ipHostInfo = Dns.Resolve(baxiAgentIpAddress);
            IPAddress _ipAddress = _ipHostInfo.AddressList[0];
            _remoteEP = new IPEndPoint(_ipAddress, port);
            _socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
        }

        /// <summary>
        /// This method tries to connect to the remote endpoint. 
        /// If the connection cannot be established the exception will be logged in the console output and tries to connect 
        /// again maximum 2 times.
        /// </summary>
        public async Task Connect()
        {
            var count = 0;
            while (!_socket.Connected && count < _retryConnectCount)
            {
                try
                {
                    var result = _socket.BeginConnect(_remoteEP, null, null);
                    bool success = result.AsyncWaitHandle.WaitOne(5000, true);
                    if (success && _socket.Connected)                    
                    {
                        _socket.EndConnect(result);
                    }
                    else
                    {
                        _socket.Close();
                        throw new SocketException(); // Connection timed out.
                    }
           
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message + Environment.NewLine);
                    Console.WriteLine(e.StackTrace + Environment.NewLine);                
                    Thread.Sleep(1000);
                }
                finally
                {
                    count++;
                }
            }
        }
   

        /// <summary>
        /// The SendString method builds the message that will be sent through the socket to the Baxi Agent. 
        /// It converts a given string into an array of bytes, reserving the first 4 bytes for the length of the message. 
        /// After that the message is sent to Baxi Agent.
        /// </summary>
        /// <param name="msg">The string contains the message that needs to be sent to Baxi Agent.</param>
        public void SendString(string msg)
        {
            // The first 4 bytes are reserved for the length of the message.
            var msgToSend = new byte[msg.Length + 4];

            // Define the length of the message.
            var lengthInBytes = GetMessageLengthInBytes(msg.Length);
            Array.Copy(lengthInBytes, msgToSend, lengthInBytes.Length);

            // Convert the string into array.
            var msgBytes = Encoding.ASCII.GetBytes(msg);
            Array.Copy(msgBytes, 0, msgToSend, lengthInBytes.Length, msgBytes.Length);

            SendMessage(msgToSend);
        }

        /// <summary>
        /// This method sends a given array of bytes to the connected socket.
        /// If no bytes could be sent to the socket, it means that the message could not be sent.
        /// </summary>
        /// <param name="msg">The message that will be sent.</param>
        /// <returns>Returns the number of bytes that were sent to the socket.</returns>
        public int SendMessage(byte[] msg)
        {
            var bytesSent = 0;
            try
            {
                bytesSent = _socket.Send(msg);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message + Environment.NewLine);
                Console.WriteLine(e.StackTrace + Environment.NewLine);
            }
            if (bytesSent <= 0)
            {
                // If the message could not be sent, then close the connection.
                if (StopSendingEmptyMessage != null)
                    StopSendingEmptyMessage(null, null);
                CloseConnection();

                Console.WriteLine(SocketError.ConnectionAborted.ToString() + Environment.NewLine);
            }
            return bytesSent;
        }

        /// <summary>
        /// The ReceiveMessage method tries to receive a message from the connected socket into a buffer.
        /// </summary>
        /// <returns>Returns the array of bytes received on the socket.</returns>
        public byte[] ReceiveMessage()
        {
            var nrBytesReceived = 0;
            var received = new byte[nrBytesReceived];

            try
            {
                _socket.ReceiveTimeout = 12000;
                nrBytesReceived = _socket.Receive(_bytesReceived);
                received = new byte[nrBytesReceived];
                Array.Copy(_bytesReceived, received, nrBytesReceived);
                return received;
            }

            catch (SocketException e)
            {
                if (StopSendingEmptyMessage != null)
                    StopSendingEmptyMessage(null, null);
                CloseConnection();
                Thread.Sleep(1000);

                Console.WriteLine(e.Message + Environment.NewLine);
                Console.WriteLine(e.StackTrace + Environment.NewLine);
                return new byte[0];
            }
            catch (Exception)
            {
                return new byte[0];
            }
        }

        /// <summary>
        /// This method closes the socket connection
        /// </summary>
        public void CloseConnection()
        {
            try
            {
                if (_socket.Connected)
                {
                    _socket.Disconnect(false);
                    _socket.Shutdown(SocketShutdown.Both);
                    _socket.Close();
                }
            }
            catch (Exception e)
            {

                Console.WriteLine(e.Message + Environment.NewLine);
                Console.WriteLine(e.StackTrace + Environment.NewLine);
            }
        }

        /// <summary>
        /// This method returns an array of 4 bytes converted into an integer number.
        /// </summary>
        /// <param name="prefix">The array of bytes that will be converted into integer.</param>
        private int GetMessageLengthInInt(byte[] prefix)
        {
            if (BitConverter.IsLittleEndian)
                Array.Reverse(prefix);
            var msgLen = BitConverter.ToInt32(prefix, 0);
            return msgLen;
        }

        /// <summary>
        /// This method returns the length of a string converted into an array of 4 bytes.
        /// </summary>
        /// <param name="length">The lenght of the string.</param>
        public static byte[] GetMessageLengthInBytes(int length)
        {
            var bytes = BitConverter.GetBytes(length);
            if (BitConverter.IsLittleEndian)
                Array.Reverse(bytes);
            return bytes; 
        }
    }
}
