﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;
using System.Configuration;
using System.Threading;
using WebSocketSharp;
using System.Net.Http;
using Baxi.NewtonSoft.Json.Linq;
using Baxi.NewtonSoft.Json;

namespace XmlProtocolSampleApp
{
    public class ClientWebsocketIO : XmlProtocolSampleApp.XmlProtocolController.SocketIO
    {

        private const int _retryConnectCount = 2;

        // a semaphore system to prevent race condition when handling the received buffer
        private Object bufferLock = new Object();

        private byte[] _bytesReceived;

        // websocket client for the persistent communication
        private WebSocket _socket;

        // authentication token for websocket communication
        private string _jsonWebToken;

        private string _baxiAgentAddress;
        private int _baxiAgentPort;
        private string _username;
        private string _password;

        public bool IsConnected { get { return _socket.IsAlive; } }

        /// <summary>
        /// Eventhandler for stopping keep-alive messages from being sent.
        /// </summary>
        public EventHandler StopSendingEmptyMessage;


        public class LoginRequest
        {
            [JsonProperty("login")]
            public Credentials LoginCredentials { get; set; }
        }

        public class Credentials
        {
            [JsonProperty("username")]
            public string Username { get; set; }

            [JsonProperty("password")]
            public string Password { get; set; }
        }

        public class LoginResponse
        {
            [JsonProperty("token")]
            public string jsonWebToken { get; set; }

            [JsonProperty("username")]
            public string username { get; set; }

            [JsonProperty("terminals")]
            public string[] terminals { get; set; }

        }



        public ClientWebsocketIO(string serverAddress, int serverPort, string terminalId, string username, string password)
        {
            _baxiAgentAddress = serverAddress;
            _baxiAgentPort = serverPort;
            _username = username;
            _password = password;

            _bytesReceived = new byte[0];                        
        }


        private async Task Setup()
        {
            /// To disable HTTPS certificate validation for self-signed certificates when
            /// using a development/test server.
            /// NOTE: do NOT use in your own production application.
            ServicePointManager.ServerCertificateValidationCallback +=
                (sender, cert, chain, sslPolicyErrors) => true;

            _jsonWebToken = await FetchJsonWebToken();
            SetUpWebSocketConnection();            
        }

        private void SetUpWebSocketConnection()
        {            
            _baxiAgentAddress = "wss://" + _baxiAgentAddress + ":" + _baxiAgentPort + "/ws/xml?auth_token=" + _jsonWebToken;

            _socket = new WebSocket(_baxiAgentAddress);
            _socket.OnMessage += HandleReceivedMessage;
        }

        /// <summary>
        /// Fetching JWT (json web token) from RESTful authentication API
        /// </summary>
        private async Task<string> FetchJsonWebToken()
        {

            var payload = new LoginRequest
            {
                LoginCredentials = new Credentials
                {
                    Username = _username,
                    Password = _password
                }
            };
                
            var stringPayload = JsonConvert.SerializeObject(payload);
            var httpContent = new StringContent(stringPayload, Encoding.UTF8, "application/json");
            
            // HTTP post request to login API endpoint
            try
            {
                var httpResponse = await new HttpClient().PostAsync("https://" + _baxiAgentAddress + ":" + _baxiAgentPort + "/api/login", httpContent);

                if (httpResponse.Content != null)
                {
                    var responseContent = await httpResponse.Content.ReadAsStringAsync();
                    LoginResponse response = JsonConvert.DeserializeObject<LoginResponse>(responseContent);
                    return response.jsonWebToken;
                }

            }
            catch (Exception e)
            {
                Console.WriteLine("JWT obtaining failed: " + e.ToString());
                return "";
            }

            return "";                        
        }


        private void HandleReceivedMessage(object sender, MessageEventArgs e)
        {

            /// we will use a semaphore-like system to prevent race condition between the event handler
            /// and ReceiveMessage() method. The original implementation was based on polling while the
            /// WebSocket library is event-driven.
            
            lock (bufferLock)
            {
                var newReceiveBuffer = new byte[_bytesReceived.Length + e.RawData.Length];

                Buffer.BlockCopy(_bytesReceived, 0, newReceiveBuffer, 0, _bytesReceived.Length);
                Buffer.BlockCopy(e.RawData, 0, newReceiveBuffer, _bytesReceived.Length, e.RawData.Length);

                _bytesReceived = newReceiveBuffer;
            }

        }

        /// <summary>
        /// This method tries to connect to the remote endpoint. 
        /// </summary>
        public async Task Connect()
        {
            await Setup();
            _socket.Connect();
        }


        /// <summary>
        /// The SendString method builds the message that will be sent through the socket to the Baxi Agent. 
        /// It converts a given string into an array of bytes, reserving the first 4 bytes for the length of the message. 
        /// After that the message is sent to Baxi Agent.
        /// </summary>
        /// <param name="msg">The string contains the message that needs to be sent to Baxi Agent.</param>
        public void SendString(string msg)
        {

            // The first 4 bytes are reserved for the length of the message.
            var msgToSend = new byte[msg.Length + 4];

            // Define the length of the message.
            var lengthInBytes = GetMessageLengthInBytes(msg.Length);
            Array.Copy(lengthInBytes, msgToSend, lengthInBytes.Length);

            // Convert the string into array.
            var msgBytes = Encoding.ASCII.GetBytes(msg);
            Array.Copy(msgBytes, 0, msgToSend, lengthInBytes.Length, msgBytes.Length);
            
            SendMessage(msgToSend);
            
        }

        /// <summary>
        /// This method sends a given array of bytes to the connected socket.
        /// If no bytes could be sent to the socket, it means that the message could not be sent.
        /// </summary>
        /// <param name="msg">The message that will be sent.</param>
        /// <returns>Returns the number of bytes that were sent to the socket.</returns>
        public int SendMessage(byte[] msg)
        {
            _socket.Send(msg);
            return msg.Length;
        }

        /// <summary>
        /// The ReceiveMessage polls the receive buffer
        /// </summary>
        /// <returns>Returns the array of bytes received on the socket.</returns>
        public byte[] ReceiveMessage()
        {

            lock (bufferLock) {

                var received = new byte[_bytesReceived.Length];
                System.Buffer.BlockCopy(_bytesReceived, 0, received, 0, _bytesReceived.Length);

                _bytesReceived = new byte[0];

                return received;
            }

        }

        /// <summary>
        /// This method closes the socket connection
        /// </summary>
        public void CloseConnection()
        {
            _socket.Close(WebSocketSharp.CloseStatusCode.Normal, "Normal closure");
        }

        /// <summary>
        /// This method returns an array of 4 bytes converted into an integer number.
        /// </summary>
        /// <param name="prefix">The array of bytes that will be converted into integer.</param>
        private int GetMessageLengthInInt(byte[] prefix)
        {
            if (BitConverter.IsLittleEndian)
                Array.Reverse(prefix);
            var msgLen = BitConverter.ToInt32(prefix, 0);
            return msgLen;
        }

        /// <summary>
        /// This method returns the length of a string converted into an array of 4 bytes.
        /// </summary>
        /// <param name="length">The lenght of the string.</param>
        public static byte[] GetMessageLengthInBytes(int length)
        {
            var bytes = BitConverter.GetBytes(length);
            if (BitConverter.IsLittleEndian)
                Array.Reverse(bytes);
            return bytes;
        }
    }
}
