﻿using Baxi.NewtonSoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Xml;
using XmlProtocolSampleApp.DTO;
using Baxi.NewtonSoft.Json.Linq;
using Baxi.NewtonSoft.Json;
using System.Threading.Tasks;

namespace XmlProtocolSampleApp
{
    /// <summary>
    /// This class is the controller class for the SampleEcr GUI.
    /// </summary>
    public class XmlProtocolController
    {
        /// <summary>
        /// Instance for the socket communication.
        /// </summary>
        private SocketIO _socket;

        public interface SocketIO
        {
            Task Connect();
            void SendString(string msg);
            int SendMessage(byte[] msg);
            byte[] ReceiveMessage();
            void CloseConnection();
            bool IsConnected { get; }
        }

        public delegate void LocalModeInfoReceivedEvent(string info, bool append = true);
        public LocalModeInfoReceivedEvent OnLocalModeReceived;
        public LocalModeInfoReceivedEvent OnTerminalReadyReceived;
        public LocalModeInfoReceivedEvent OnJsonReceived;

        public delegate void DisplayTextInfoReceivedEvent(string info, bool append = true);
        public DisplayTextInfoReceivedEvent OnDisplayTextReceived;
        public DisplayTextInfoReceivedEvent OnMethodRejectedReceived;
        public DisplayTextInfoReceivedEvent OnErrorReceived;

        public delegate void GetBaxiPropertiesEvent(Dictionary<string, string> values);
        public GetBaxiPropertiesEvent OnGetProperties;

        public EventHandler OnBaxiPropertiesSet;

        public delegate void PrintTextInfoReceivedevent(string info, bool append = true);
        public PrintTextInfoReceivedevent OnPrintTextReceived;

        public delegate void OperatorConfirmationRequestReceivedEvent(int messageType, string info);
        public OperatorConfirmationRequestReceivedEvent OnPinBypassRequestReceived;


        private Thread _receiveThread;
        private Thread _keepAliveThread;
        private string _terminalId;
        private string _ecrId;

        // public XmlProtocolController(string termId, string ecrId, string agentIP, int agentPort, string cloudUsername, string cloudPassword)
        public XmlProtocolController(XmlProtocolSampleApp.SampleEcr.TerminalConnectionConfiguration config)
        {
            _terminalId = config.TerminalId;
            _ecrId = config.EcrId;
            _socket = config.ConnectionType == SampleEcr.TerminalConnectionType.BAXIAGENT ? (SocketIO) new ClientSocketIO(config.AgentIP, config.AgentPort) : new ClientWebsocketIO(config.AgentIP, config.AgentPort, config.TerminalId, config.Username, config.Password);
        }

        /// <summary>
        /// Opens the connection on the socket configured in the constructor. 
        /// After the socket is connected, an "Open" request is sent to Baxi Agent. 
        /// A new thread is created for receiving messages from Baxi Agent and another one for sending Keep alive messages 
        /// to Baxi Agent every 5 seconds.
        /// </summary>
        public async void OpenConnection()
        {
            await _socket.Connect(); // wait for the socket to connect before sending any messages.
            var doc = new XmlDocument();
            doc.LoadXml(XmlTemplates.Open);
            var termIdAttr = doc.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null)
                termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = doc.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null)
                ecrIdAttr.InnerText = _ecrId;
            _socket.SendString(doc.InnerXml);

            _receiveThread = new Thread(ReceiveMessages);
            _receiveThread.IsBackground = true;
            _receiveThread.Priority = ThreadPriority.Normal;
            _receiveThread.Start();

            _keepAliveThread = new Thread(KeepAlive);
            //kill the threads when main thread ends
            _keepAliveThread.IsBackground = true;
            _keepAliveThread.Priority = ThreadPriority.Normal;
            _keepAliveThread.Start();
        }

        /// <summary>
        /// Method to receive messages from Baxi Agent.
        /// </summary>
        public void ReceiveMessages()
        {
            while (_socket.IsConnected)
            {
                var rec = _socket.ReceiveMessage();
                HandleReceivedMessages(rec);
            }
        }

        /// <summary>
        /// HandleReceivedMessages parses the byte array received from Baxi Agent and transforms it into an XML document.
        /// Based on the response contained in the XML, we fire the proper action in the UI.
        /// </summary>
        /// <param name="rec">The byte array received from Baxi Agent</param>
        private void HandleReceivedMessages(byte[] rec)
        {
            var index = 0;
            while (index < rec.Length)
            {
                if (rec.Length > 0)
                {
                    // get the length of the message
                    var lenBytes = new byte[4];
                    Array.Copy(rec, index, lenBytes, 0, 4);
                    index += 4;
                    if (BitConverter.IsLittleEndian)
                        Array.Reverse(lenBytes);
                    var msgLen = BitConverter.ToInt32(lenBytes, 0);

                    if (msgLen > 0)
                    {
                        // get the message as a string
                        var msgRec = new byte[msgLen];
                        Array.Copy(rec, index, msgRec, 0, msgLen);
                        index += msgLen;
                        var msgStr = Encoding.UTF8.GetString(msgRec);

                        // create an xml document based on the response string
                        XmlDocument xml = new XmlDocument();
                        xml.LoadXml(msgStr);

                        // Retrieve the actual response.
                        // Since every NetsResponse XML tag has a "MessageHeader" child node and the next child node is the response itself, 
                        // we can relate to it as the "MessageHeader" node's sibling.
                        var response = xml.GetElementsByTagName("MessageHeader")[0].NextSibling;

                        // we trigger an event based on the response we retrieve from the response XML
                        switch (response.Name)
                        {
                            case "Dfs13LocalMode":
                                if (OnLocalModeReceived != null)
                                {
                                    var sb = new StringBuilder();
                                    sb.AppendLine("Local mode: ");

                                    foreach (XmlNode item in response.ChildNodes)
                                        sb.AppendFormat(@"{0}: {1} {2}", item.Name, item.InnerText, Environment.NewLine);

                                    sb.AppendLine();
                                    OnLocalModeReceived(sb.ToString());
                                }
                                break;
                            case "Dfs13TerminalReady":
                                if (OnTerminalReadyReceived != null)
                                {
                                    var sb = new StringBuilder();
                                    sb.AppendLine("TerminalReady event received.");
                                    sb.AppendLine();
                                    OnTerminalReadyReceived(sb.ToString());
                                }
                                break;
                            case "Dfs13DisplayText":
                                if (OnDisplayTextReceived != null)
                                {
                                    var sb = new StringBuilder();
                                    sb.AppendLine("DisplayText: ");

                                    foreach (XmlNode item in response.ChildNodes)
                                        sb.AppendFormat(@"{0} ", item.InnerText);

                                    sb.AppendLine();
                                    sb.AppendLine();
                                    OnDisplayTextReceived(sb.ToString());
                                }
                                break;
                            case "Dfs13PrintText":
                                if (OnPrintTextReceived != null)
                                {
                                    var sb = new StringBuilder();
                                    sb.AppendLine("PrintText: ");

                                    foreach (XmlNode item in response.ChildNodes)
                                        sb.AppendFormat(@"{0} ", item.InnerText);

                                    sb.AppendLine();
                                    sb.AppendLine();
                                    OnPrintTextReceived(sb.ToString());
                                }
                                break;
                            case "Dfs13JsonReceived":
                                if (OnJsonReceived != null)
                                {
                                    var sb = new StringBuilder();
                                    sb.AppendLine("Json Received: ");

                                    foreach (XmlNode item in response.ChildNodes)
                                    {
                                        sb.AppendFormat(@"{0}: {1} {2}", item.Name, item.InnerText, Environment.NewLine);
                                        if (item.Name.Equals("Data")) ProcessJsonMessage(item.InnerText);
                                    }                                        

                                    sb.AppendLine();
                                    sb.AppendLine();                                                                      

                                    OnJsonReceived(sb.ToString());
                                }
                                break;
                            case "Dfs13Error":
                                if (OnErrorReceived != null)
                                {
                                    var sb = new StringBuilder();
                                    sb.AppendLine("Error: ");
                                    foreach (XmlNode item in response.ChildNodes)
                                        sb.AppendFormat(@"{0} ", item.InnerText);
                                    sb.AppendLine();
                                    sb.AppendLine();
                                    OnErrorReceived(sb.ToString());
                                }
                                break;
                            case "MethodRejected":
                                if (OnMethodRejectedReceived != null)
                                {
                                    var sb = new StringBuilder();
                                    sb.AppendLine("MethodRejected: ");
                                    foreach (XmlNode item in response.ChildNodes)
                                        sb.AppendFormat(@"{0} ", item.InnerText);
                                    sb.AppendLine();
                                    sb.AppendLine();
                                    OnMethodRejectedReceived(sb.ToString());
                                }
                                break;
                            case "KeepAlive":
                                Console.WriteLine("Keep Alive message received");
                                break;

                            case "BaxiProperties":
                                if (OnGetProperties != null)
                                {
                                    var values = new Dictionary<string, string>();
                                    foreach (XmlNode item in response.ChildNodes)
                                    {
                                        var value = item.InnerText;
                                        if (item.Name == "Version" || item.Name == "TerminalDeviceData_TLD")
                                        {
                                            value = StringToSeparators(item.InnerText);
                                        }
                                        values.Add(item.Name, value);
                                    }
                                    OnGetProperties(values);
                                }
                                break;

                            case "SetBaxiProperties":
                                if (OnBaxiPropertiesSet != null)
                                    OnBaxiPropertiesSet(null, null);
                                break;
                            default:
                                break;
                        }
                    }
                }
            }
        }


        public void KeepAlive()
        {
            while (_socket.IsConnected)
            {
                Thread.Sleep(5000);
                var doc = new XmlDataDocument();
                doc.LoadXml(XmlTemplates.KeepAlive);
                var termIdAttr = doc.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
                if (termIdAttr != null)
                    termIdAttr.InnerText = _terminalId;
                var ecrIdAttr = doc.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
                if (ecrIdAttr != null)
                    ecrIdAttr.InnerText = _ecrId;
                _socket.SendString(doc.InnerXml);
            }

        }
        /// <summary>
        /// This method sends the TransferAmount request to Baxi Agent
        /// </summary>
        /// <param name="amount">Specifies the amount that needs to be paid</param>
        /// <param name="transactiontype">Specifies the transactiontype</param>
        public void CardPayment(string amount, string transactiontype)
        {
            var transferAmnountRequest = new XmlDocument();
            transferAmnountRequest.LoadXml(XmlTemplates.TransferAmount);
            SetNodeValue(transferAmnountRequest,"NetsRequest/Dfs13TransferAmount/TransactionType",transactiontype);

            var termIdAttr = transferAmnountRequest.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null)
                termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = transferAmnountRequest.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null)
                ecrIdAttr.InnerText = _ecrId;
            var amountNode = transferAmnountRequest.SelectSingleNode("NetsRequest/Dfs13TransferAmount/Amount1");
            if (amountNode != null)
                amountNode.InnerText = amount;
            _socket.SendString(transferAmnountRequest.InnerXml.ToString());
        }

        static public void SetNodeValue(XmlDocument xml, string xpathNodeName, string value)
        {
            XmlNode node = null;
            node = xml.SelectSingleNode(xpathNodeName); //Takes the FIRST match
            if (node != null)
            {
                node.InnerText = value;
            }
        }

        public void TransferCardData(string data)
        {
            var transferCardDataRequest = new XmlDocument();
            transferCardDataRequest.LoadXml(XmlTemplates.TransferCardData);
            SetNodeValue(transferCardDataRequest, "NetsRequest/Dfs13TransferCardData/Track2", data);
            SetNodeValue(transferCardDataRequest, "NetsRequest/Dfs13TransferCardData/Origin", "50");

            var termIdAttr = transferCardDataRequest.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null)
                termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = transferCardDataRequest.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null)
                ecrIdAttr.InnerText = _ecrId;
 
            _socket.SendString(transferCardDataRequest.InnerXml.ToString());
        }

        /// <summary>
        /// This method sends Admin Cancel request to Baxi Agent
        /// </summary>
        public void AdminCancel()
        {
            /// 0x3132 - admin code for cancel
            var adminCancel = BuildAdminCommand((int)0x3132);
            _socket.SendString(adminCancel.InnerXml.ToString());
        }

        /// <summary>
        /// EndOfDay method sends Admin reconciliation request to Baxi Agent
        /// </summary>
        public void EndOfDay()
        {
            /// 0x31302 - admin code for End of Day
            var adminEOD = BuildAdminCommand((int)0x3130);
            _socket.SendString(adminEOD.InnerXml.ToString());
        }

        /// <summary>
        /// This method sends Admin Get last financial receipt request to Baxi Agent
        /// </summary>
        public void LastFinancialReceipt()
        {
            /// 0x313C - admin command for Last Financial Receipt
            var adminLFR = BuildAdminCommand((int)0x313C);
            _socket.SendString(adminLFR.InnerXml.ToString());
        }

        /// <summary>
        /// Helper method for parsing JSON formatted messages and passing appropriate events
        /// to UI layer.
        /// </summary>
        /// <param name="messageContents">A string containing the raw JSON data</param>
        private void ProcessJsonMessage(string messageContents)
        {
            dynamic json = JObject.Parse(messageContents);

            if (json.confirm != null)
            {
                int id = json.confirm.id;

                switch (id)
                {
                    case 1: // PIN bypass
                        OnPinBypassRequestReceived(id, "PIN Bypass requested.");
                        break;
                    default:
                        break;
                }
            }
        }

        /// <summary>
        /// Send a JSON-formatted string to BAXI
        /// </summary>
        /// <param name="jsonMessage">JSON-formatted message</param>
        public void SendJson(string jsonMessage)
        {
            var sendJsonMessage = new XmlDocument();
            sendJsonMessage.LoadXml(XmlTemplates.SendJson);

            var termIdAttr = sendJsonMessage.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null) termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = sendJsonMessage.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null) ecrIdAttr.InnerText = _ecrId;

            var data = sendJsonMessage.SelectSingleNode("NetsRequest/Dfs13SendJson/Data");
            if (data != null) data.InnerText = jsonMessage;

            _socket.SendString(sendJsonMessage.InnerXml.ToString());
        }


        /// <summary>
        /// Builds the Administration request based on given administration code
        /// </summary>
        private XmlDocument BuildAdminCommand(int admCode)
        {
            var adminCancel = new XmlDocument();
            adminCancel.LoadXml(XmlTemplates.Administration);
            var termIdAttr = adminCancel.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null)
                termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = adminCancel.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null)
                ecrIdAttr.InnerText = _ecrId;
            var admCodeNode = adminCancel.SelectSingleNode("NetsRequest/Dfs13Administration/AdmCode");
            if (admCodeNode != null)
                admCodeNode.InnerText = admCode.ToString();
            return adminCancel;
        }

        /// <summary>
        /// This method sends a Json request to the Baxi Agent, querying all tags
        /// </summary>
        public void GetCardInfo()
        {
            var cardInfo = new XmlDocument();
            cardInfo.LoadXml(XmlTemplates.SendJson);
            var termIdAttr = cardInfo.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null)
                termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = cardInfo.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null)
                ecrIdAttr.InnerText = _ecrId;
            var data = cardInfo.SelectSingleNode("NetsRequest/Dfs13SendJson/Data");
            if (data != null)
                // Json for querying all tags
                data.InnerText =
@"{
    ""cardinfo"": {
        ""ver"": ""1.01"",
        ""alltags"": ""?""
    },
    ""o"": {
        ""ver"": ""1.00"",
        ""merch"": 733300
    }
}";

            _socket.SendString(cardInfo.InnerXml.ToString());
        }

        public void GetRWProperties()
        {
            var getBaxiProperties = new XmlDocument();
            getBaxiProperties.LoadXml(XmlTemplates.BaxiProperties);
            var termIdAttr = getBaxiProperties.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null)
                termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = getBaxiProperties.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null)
                ecrIdAttr.InnerText = _ecrId;
            _socket.SendString(getBaxiProperties.InnerXml.ToString());
        }


        public void CloseConnection()
        {
            _socket.CloseConnection();
        }

        internal void SetBaxiProperties(Dictionary<string, string> changedValues)
        {
            var setWProperties = new XmlDocument();
            setWProperties.LoadXml(XmlTemplates.SetBaxiProperties);
            var termIdAttr = setWProperties.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null)
                termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = setWProperties.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null)
                ecrIdAttr.InnerText = _ecrId;

            foreach (var item in changedValues)
            {
                XmlNode node = null;
                node = setWProperties.SelectSingleNode(XmlTemplates.XlsSetBaxiProperties);


                XmlNode root = setWProperties.DocumentElement;

                //Create a new node.
                XmlElement elem = setWProperties.CreateElement(item.Key);
                elem.InnerText = item.Value;

                //Add the node to the document.
                node.AppendChild(elem);

            }

            _socket.SendString(setWProperties.InnerXml.ToString());
        }


        internal static string StringToSeparators(string data)
        {
            string s = data;
            if (s == null)
            {
                s = "";
            }
            if (s != "")
            {
                //Used for TLD
                s = s.Replace(XmlFormat.UsStr, Convert.ToChar(XmlFormat.US).ToString());
                s = s.Replace(XmlFormat.RsStr, Convert.ToChar(XmlFormat.RS).ToString());
                s = s.Replace(XmlFormat.FsStr, Convert.ToChar(XmlFormat.FS).ToString());
                s = s.Replace(XmlFormat.NpStr, Convert.ToChar(XmlFormat.NP).ToString());

                //Used for PrintText
                s = s.Replace(XmlFormat.SoStr, Convert.ToChar(XmlFormat.SO).ToString());

                //Used for XML handling of <> in Baxi.Version <C>
                s = s.Replace(XmlFormat.LessStr, Convert.ToChar(XmlFormat.LESS).ToString());
                s = s.Replace(XmlFormat.MoreStr, Convert.ToChar(XmlFormat.MORE).ToString());
            }
            return s;
        }

        public string BuildStoreboxReceipt(List<Tuple<string, int, int>> products)
        {
            var receipt = new StoreBoxReceipt();

            var userId = new StoreBoxUserId();
            userId.type = StoreBoxEnums.UserIdType.pan;
            userId.value = "a952694b1b0546495ea96a213539ed6b355fd236444351c6ab79b19e3344628b";
            receipt.userId = userId;

            var receiptId = new StoreBoxReceiptId();
            receiptId.type = "Receipt";
            receiptId.value = "123";
            receipt.receiptId = receiptId;

            receipt.merchantId = "ABC";

            var store = new StoreBoxStore();
            store.storeName = "ABC Store";
            store.storeId = "17226";
            store.storeAddressLine1 = "Algade 12";
            store.storeZipCode = "9900";
            store.storeCity = "Aalborg";
            store.storeCountry = "Danmark";
            receipt.store = store;

            int hours = TimeZoneInfo.Local.BaseUtcOffset.Hours;
            int minutes = TimeZoneInfo.Local.BaseUtcOffset.Minutes;
            string offset = string.Format("{0}{1}:{2}", ((hours > 0) ? "+" : ""), hours.ToString("00"), minutes.ToString("00"));
            string isoformat = DateTime.Now.ToString("s") + offset;
            receipt.orderDate = isoformat;

            receipt.orderNumber = "25";

            var index = 0;
            decimal total = 0;
            decimal totalVat = 0;
            receipt.orderLines = new StoreBoxOrderLine[products.Count];
            foreach (var item in products)
            {
                var orderline = new StoreBoxOrderLine();
                orderline.number = (index + 1).ToString();
                orderline.category = "Products";
                orderline.name = item.Item1;
                orderline.text = item.Item1;
                var itemPrice = new StoreBoxItemPrice();
                itemPrice.value = (decimal)(item.Item2 * 0.01);
                itemPrice.vatPercentage = (decimal)25.00;
                itemPrice.vat = decimal.Round((itemPrice.value * itemPrice.vatPercentage / 100), 2, MidpointRounding.AwayFromZero);
                orderline.itemPrice = itemPrice;

                orderline.count = (decimal)item.Item3;

                var totalPrice = new StoreBoxTotalPrice();
                totalPrice.value = (decimal)(itemPrice.value * orderline.count);
                totalPrice.vat = (decimal)(itemPrice.vat * orderline.count);
                orderline.totalPrice = totalPrice;
                receipt.orderLines[index] = orderline;
                total += totalPrice.value;
                totalVat += totalPrice.vat;
                index++;
            }

            var totalOrderPrice = new StoreBoxTotalOrderPrice();
            totalOrderPrice.value = total;
            totalOrderPrice.vat = totalVat;
            totalOrderPrice.currency = StoreBoxEnums.Currency.DKK;
            receipt.totalOrderPrice = totalOrderPrice;

            var payment = new StoreBoxPayment();
            payment.paymentType = StoreBoxEnums.PaymentType.paymentcard;
            payment.paymentCardType = StoreBoxEnums.PaymentCardType.visa;
            var price = new StoreBoxPrice();
            price.value = (decimal)total;
            price.currency = StoreBoxEnums.Currency.DKK;
            payment.price = price;
            receipt.payments = new StoreBoxPayment[] { payment };

            var storeBoxReceipt = new StoreBox();
            storeBoxReceipt.ver = "1.00";
            storeBoxReceipt.storeBox = receipt;

            return JsonConvert.SerializeObject(storeBoxReceipt, Baxi.NewtonSoft.Json.Formatting.Indented, new JsonSerializerSettings { NullValueHandling = NullValueHandling.Ignore });
        }

        public void SendStoreBoxReceipt(String json)
        {

            var storeboxReceipt = new XmlDocument();
            storeboxReceipt.LoadXml(XmlTemplates.SendJson);
            var termIdAttr = storeboxReceipt.SelectSingleNode("NetsRequest/MessageHeader/@TerminalID");
            if (termIdAttr != null)
                termIdAttr.InnerText = _terminalId;
            var ecrIdAttr = storeboxReceipt.SelectSingleNode("NetsRequest/MessageHeader/@ECRID");
            if (ecrIdAttr != null)
                ecrIdAttr.InnerText = _ecrId;
            var data = storeboxReceipt.SelectSingleNode("NetsRequest/Dfs13SendJson/Data");
            if (data != null)
                // Json for querying all tags
                data.InnerText = json;

            _socket.SendString(storeboxReceipt.InnerXml.ToString());

            //int result = _baxi.SendJson(new SendJsonArgs(json));
            //CheckMethodRejected(result);
        }



        /// <summary>
        /// Method for constructing and sending a JSON-formatted reply to confirmation requests,
        /// such as PIN bypass request, back to BAXI.
        /// </summary>
        /// <param name="messageType">confirmation message type ID (for example 1 = PinBypass)</param>
        /// <param name="requestAccepted">true/false whether the request accepted/declined</param>
        public void SendConfirmationRequestReply(int messageType, bool requestAccepted)
        {
            ConfirmationRequestReply reply = new ConfirmationRequestReply { ver = "1.00", id = messageType, allow = (requestAccepted ? 1 : 0) };

            JObject msg = new JObject();
            msg.Add("confirm", JObject.FromObject(reply));

            SendJson(JsonConvert.SerializeObject(msg));
        }


        /// <summary>
        /// A dummy class for constructing JSON-formatted confirmation request replies 
        /// to be sent back to BAXI 
        /// </summary>
        private class ConfirmationRequestReply
        {
            public string ver { get; set; }
            public int id { get; set; }
            public int allow { get; set; }
        }




    }
}
